import logging

logger = logging.getLogger()
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    "%(asctime)s - %(process)d - %(levelname)s - %(filename)s:%(lineno)s - %(message)s"
)
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)



import argparse

import openai

openai.api_base = "http://localhost:8000"
openai.api_key = "test"
openai.debug = True


def chat():
    chat_completion = openai.ChatCompletion.create(
        model="fastchat-t5-3b-v1.0",
        messages=[
            {"role": "user", "content": "Who are you?"},
            {"role": "assistant", "content": "I am a student"},
            {"role": "user", "content": "What do you learn?"},
            {"role": "assistant", "content": "I learn math"},
            {"role": "user", "content": "Do you like english?"},
        ],
        max_tokens=100,
    )
    print(chat_completion)


def embedding():
    emb = openai.Embedding.create(
        input=["Once upon a time", "There was a frog", "Who lived in a well"],
        model="text-embedding-ada-002",
    )
    print(emb)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--chat", action="store_true")
    parser.add_argument("-e", "--embedding", action="store_true")
    args = parser.parse_args()

    if args.chat:
        chat()
    if args.embedding:
        embedding()



import asyncio
import multiprocessing as mp
import pathlib
import pickle
import socket
import struct
from typing import Any

from modelz_llm.log import logger


def run_server(path: str, barrier: mp.Barrier, cls, **kwargs):
    proc = mp.get_context("spawn").Process(
        target=Server, args=(path, barrier, cls), kwargs=kwargs, daemon=True
    )
    proc.start()
    return proc


class Server:
    def __init__(self, path: str, barrier: mp.Barrier, cls, **kwargs):
        self.path = pathlib.Path(path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        if self.path.exists():
            self.path.unlink()
        try:
            self.func = cls(**kwargs)
        finally:
            barrier.wait()
        self.run()

    def run(self):
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
            sock.bind(str(self.path))
            sock.listen()
            while True:
                conn, addr = sock.accept()
                with conn:
                    logger.info("connected to a client: '%s'", addr)
                    while True:
                        num_bytes = conn.recv(4)
                        if not num_bytes:
                            logger.info("cannot receive data")
                            break
                        num = struct.unpack("!I", num_bytes)[0]
                        data = conn.recv(num)
                        try:
                            req = pickle.loads(data)
                            resp = self.func(req)
                        except Exception as err:
                            logger.warning(err)
                            resp = err
                        data = pickle.dumps(resp, protocol=pickle.HIGHEST_PROTOCOL)
                        conn.sendall(struct.pack("!I", len(data)))
                        conn.sendall(data)

                    logger.info("closing the connection")


class Client:
    def __init__(self, path: str) -> None:
        self.path = path
        self.reader = self.writer = None

    async def request(self, req: Any) -> Any:
        if self.reader is None or self.writer is None:
            self.reader, self.writer = await asyncio.open_unix_connection(self.path)

        data = pickle.dumps(req, protocol=pickle.HIGHEST_PROTOCOL)
        self.writer.write(struct.pack("!I", len(data)))
        self.writer.write(data)
        await self.writer.drain()

        num_bytes = await self.reader.read(4)
        num = struct.unpack("!I", num_bytes)[0]
        data = await self.reader.read(num)
        return pickle.loads(data)

    async def close(self):
        self.writer.close()
        await self.writer.wait_closed()



import argparse

import uvicorn

from modelz_llm.falcon_service import build_falcon_app


def parse_argument():
    parser = argparse.ArgumentParser(
        prog="modelz-llm",
        description="Modelz Language Model Service CLI",
    )

    parser.add_argument(
        "-m",
        "--model",
        help="Language model name, for example: `facebook/dino-vitb16`",
        default="bigscience/bloomz-560m",
    )
    parser.add_argument(
        "--emb-model",
        help="Embedding model name",
        default="sentence-transformers/all-MiniLM-L6-v2",
    )
    parser.add_argument(
        "--dry-run",
        help="Dry run will only init the model without starting the server",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-d",
        "--device",
        help=(
            "Device to run the model, `auto` means it will try to use"
            " GPU if available (default=auto)"
        ),
        choices=["cpu", "cuda", "auto"],
        default="auto",
    )
    parser.add_argument(
        "--port",
        help="Port number",
        default=8000,
        type=int,
    )
    parser.add_argument(
        "--worker",
        help="Number of workers",
        default=1,
        type=int,
    )

    args = parser.parse_args()
    return args


def main():
    args = parse_argument()
    print(args)

    app = build_falcon_app(args)
    if args.dry_run:
        print("Dry run, exiting...")
        return

    uvicorn.run(
        app,
        host="0.0.0.0",
        port=args.port,
        workers=args.worker,
        access_log=False,
    )



import argparse
import multiprocessing as mp
import time

import falcon
import msgspec
from falcon.asgi import App, Request, Response
from huggingface_hub import snapshot_download
from llmspec import (
    ChatCompletionRequest,
    EmbeddingRequest,
    ErrorResponse,
    PromptCompletionRequest,
)

from modelz_llm.emb import Emb
from modelz_llm.log import logger
from modelz_llm.model import LLM
from modelz_llm.uds import Client, run_server

UNIX_DOMAIN_SOCKET_PATH = "/tmp/modelz_llm/{}.sock"


class Ping:
    async def on_get(self, req: Request, resp: Response):
        resp.content_type = falcon.MEDIA_TEXT
        resp.text = "Modelz LLM service"


class ChatCompletions:
    def __init__(self, client: Client) -> None:
        self.client = client

    async def on_post(self, req: Request, resp: Response):
        buf = await req.stream.readall()
        try:
            chat_req = ChatCompletionRequest.from_bytes(buf=buf)
        except msgspec.ValidationError as err:
            logger.info(f"Failed to parse request: {err}")
            # return 400 otherwise the client will retry
            resp.status = falcon.HTTP_400
            resp.data = ErrorResponse.from_validation_err(err, str(buf)).to_json()
            return

        comp = await self.client.request(chat_req)
        logger.info(comp)
        if isinstance(comp, Exception):
            resp.status = falcon.HTTP_500
            resp.data = ErrorResponse.from_validation_err(
                comp, "internal error"
            ).to_json()
            return
        resp.data = comp.to_json()


class Completions:
    def __init__(self, client: Client) -> None:
        self.client = client

    async def on_post(self, req: Request, resp: Response):
        buf = await req.stream.readall()
        try:
            prompt_req = PromptCompletionRequest.from_bytes(buf=buf)
        except msgspec.ValidationError as err:
            logger.info(f"Failed to parse request: {err}")
            # return 400 otherwise the client will retry
            resp.status = falcon.HTTP_400
            resp.data = ErrorResponse.from_validation_err(err, str(buf)).to_json()
            return

        comp = await self.client.request(prompt_req)
        logger.info(comp)
        if isinstance(comp, Exception):
            resp.status = falcon.HTTP_500
            resp.data = ErrorResponse.from_validation_err(
                comp, "internal error"
            ).to_json()
            return
        resp.data = comp.to_json()


class Embeddings:
    def __init__(self, client: Client) -> None:
        self.client = client

    async def on_post(self, req: Request, resp: Response, engine: str = ""):
        if engine:
            logger.info("received emb req with engine: %s", engine)

        buf = await req.stream.readall()
        try:
            embedding_req = EmbeddingRequest.from_bytes(buf=buf)
        except msgspec.ValidationError as err:
            logger.info(f"Failed to parse request: {err}")
            resp.status = falcon.HTTP_400
            resp.data = ErrorResponse.from_validation_err(err, str(buf)).to_json()
            return

        emb = await self.client.request(embedding_req)
        if isinstance(emb, Exception):
            resp.status = falcon.HTTP_500
            resp.data = ErrorResponse.from_validation_err(
                emb, "internal error"
            ).to_json()
            return
        resp.data = emb.to_json()


def build_falcon_app(args: argparse.Namespace):
    if args.dry_run:
        snapshot_download(repo_id=args.model)
        snapshot_download(repo_id=args.emb_model)
        return

    llm_uds_path = UNIX_DOMAIN_SOCKET_PATH.format("llm")
    emb_uds_path = UNIX_DOMAIN_SOCKET_PATH.format("emb")
    barrier = mp.get_context("spawn").Barrier(3)
    model_proc = run_server(
        llm_uds_path, barrier, LLM, model_name=args.model, device=args.device
    )
    emb_proc = run_server(
        emb_uds_path, barrier, Emb, model_name=args.emb_model, device=args.device
    )
    barrier.wait()
    # wait to detect the proc exitcode
    time.sleep(1)
    if any(proc.exitcode is not None for proc in [model_proc, emb_proc]):
        raise RuntimeError("failed to start the service")

    llm_client = Client(llm_uds_path)
    completion = Completions(llm_client)
    chat_completion = ChatCompletions(llm_client)
    emb_client = Client(emb_uds_path)
    embeddings = Embeddings(emb_client)

    app = App()
    app.add_route("/", Ping())
    app.add_route("/completions", completion)
    app.add_route("/chat/completions", chat_completion)
    app.add_route("/embeddings", embeddings)
    app.add_route("/engines/{engine}/embeddings", embeddings)

    # refer to https://platform.openai.com/docs/api-reference/chat
    # make it fully compatible with the current OpenAI API endpoints
    app.add_route("/v1/completions", completion)
    app.add_route("/v1/chat/completions", chat_completion)
    app.add_route("/v1/embeddings", embeddings)
    app.add_route("/v1/engines/{engine}/embeddings", embeddings)
    return app



import gc
from typing import Iterable, List, Union

import torch
import transformers
from llmspec import (
    ChatCompletionRequest,
    ChatResponse,
    CompletionResponse,
    LanguageModels,
    PromptCompletionRequest,
)

from modelz_llm.log import logger
from modelz_llm.utils import (
    MIN_TEMPERATURE,
    MIN_TOP_P,
    partial_stop,
    prepare_logits_processor,
)

CONTEXT_LEN = 2048


class LLM:
    def __init__(self, model_name: str, device: str) -> None:
        self.model_name = model_name
        self.model_spec = LanguageModels.find(model_name).value
        tokenizer_cls = getattr(transformers, self.model_spec.tokenizer_cls)
        model_cls = getattr(transformers, self.model_spec.transformer_model_cls)

        logger.info(
            "loading model and embedding: %s(%s) %s(%s)",
            model_cls,
            model_name,
            tokenizer_cls,
            model_name,
        )
        self.tokenizer = tokenizer_cls.from_pretrained(
            model_name,
            trust_remote_code=True,
            low_cpu_mem_usage=self.model_spec.low_cpu_mem_usage,
        )
        self.model = model_cls.from_pretrained(
            model_name,
            trust_remote_code=True,
            low_cpu_mem_usage=self.model_spec.low_cpu_mem_usage,
        )
        if device == "auto":
            self.device = (
                torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
            )
        else:
            self.device = device

        try:
            if self.device == "cpu":
                self.model = self.model.float()
            else:
                self.model = self.model.half()
        except Exception as err:
            logger.debug("failed to convert the model: %s", err)

        self.model = self.model.to(self.device)
        self.model.eval()

    def __str__(self) -> str:
        return f"LLM(model={self.model}, tokenizer={self.tokenizer})"

    def token_encode(self, text: str):
        """Encode with tokenizer."""
        tokens = self.tokenizer(text, return_tensors="pt")
        return tokens.input_ids

    def token_decode(self, token: List[int]):
        """Decode with tokenizer."""
        text = self.tokenizer.decode(token, skip_special_tokens=True)
        return text

    def generate(self, tokens, **kwargs):
        """Generate content with model."""
        inputs = tokens.to(self.device)
        outputs = self.model.generate(inputs, **kwargs).tolist()
        return outputs

    def get_prompt_from_req(
        self, req: Union[ChatCompletionRequest, PromptCompletionRequest]
    ):
        if isinstance(req, ChatCompletionRequest):
            return req.get_prompt(self.model_name)
        return req.get_prompt()

    def __call__(
        self, req: Union[ChatResponse, CompletionResponse]
    ) -> Union[ChatResponse, ChatCompletionRequest]:
        """Generate chat or completion response."""
        resp_cls = (
            ChatResponse
            if isinstance(req, ChatCompletionRequest)
            else CompletionResponse
        )
        if self.model_spec is not LanguageModels.CHAT_GLM.value:
            return list(self.step_generate(req, resp_cls=resp_cls))[0]

        tokens = self.token_encode(self.get_prompt_from_req(req))
        input_length = len(tokens[0])
        outputs = self.generate(tokens, **req.get_inference_args(self.model_name))[0]
        message = self.token_decode(outputs[input_length:])
        return resp_cls.from_message(
            message=message,
            model=self.model_name,
            finish_reason=None,
            prompt_token=input_length,
            completion_token=len(outputs) - input_length,
        )

    def step_generate(
        self,
        req: ChatCompletionRequest,
        resp_cls: Union[ChatResponse, CompletionResponse],
        echo=False,
        stream_interval=1,
    ):
        """Ref to FastChat.

        https://github.com/lm-sys/FastChat/blob/8e38141ff5dd15f3138ccfd312dd73a471e986a1/fastchat/serve/inference.py#L58
        """
        prompt = self.get_prompt_from_req(req)
        input_ids = self.token_encode(prompt)
        input_length = len(input_ids[0])
        logits_processor = prepare_logits_processor(
            req.temperature,
            req.repetition_penalty,
            req.top_p,
            req.top_k,
        )

        if not req.stop:
            stop_token_ids = []
        elif isinstance(req.stop, list):
            stop_token_ids = req.stop
        else:
            stop_token_ids = self.token_encode(req.stop).tolist()[0]
        stop_token_ids.append(self.tokenizer.eos_token_id)
        stop_token = req.stop or self.tokenizer.eos_token

        output_ids = input_ids.tolist()[0]

        is_encoder_decoder = getattr(self.model, "encoder", None) and getattr(
            self.model, "decoder", None
        )
        # encoding
        if is_encoder_decoder:
            max_src_len = CONTEXT_LEN
            input_ids = input_ids[-max_src_len:]
            encoder_output = self.model.encoder(
                input_ids=torch.as_tensor(input_ids, device=self.device)
            )[0]
            start_ids = torch.as_tensor(
                [[self.model.generation_config.decoder_start_token_id]],
                dtype=torch.int64,
                device=self.device,
            )
        else:
            max_src_len = CONTEXT_LEN - req.max_tokens - 8

        past_key_values = out = token = None
        for i in range(req.max_tokens):
            if i == 0:
                if is_encoder_decoder:
                    out = self.model.decoder(
                        input_ids=start_ids,
                        encoder_hidden_states=encoder_output,
                        use_cache=True,
                    )
                    logits = self.model.lm_head(out[0])
                else:
                    out = self.model(
                        torch.as_tensor(input_ids, device=self.device), use_cache=True
                    )
                    logits = out.logits
                past_key_values = out.past_key_values
            else:
                if is_encoder_decoder:
                    out = self.model.decoder(
                        input_ids=torch.as_tensor([[token]], device=self.device),
                        encoder_hidden_states=encoder_output,
                        use_cache=True,
                        past_key_values=past_key_values,
                    )

                    logits = self.model.lm_head(out[0])
                else:
                    out = self.model(
                        input_ids=torch.as_tensor([[token]], device=self.device),
                        use_cache=True,
                        past_key_values=past_key_values,
                    )
                    logits = out.logits
                past_key_values = out.past_key_values

            if req.repetition_penalty > 1:
                tmp_output_ids = torch.as_tensor([output_ids], device=logits.device)
            else:
                tmp_output_ids = None
            last_token_logits = logits_processor(tmp_output_ids, logits[:, -1, :])[0]

            if req.temperature < MIN_TEMPERATURE or req.top_p < MIN_TOP_P:  # greedy
                token = int(torch.argmax(last_token_logits))
            else:
                probs = torch.softmax(last_token_logits, dim=-1)
                token = int(torch.multinomial(probs, num_samples=1))

            output_ids.append(token)
            stopped = token in stop_token_ids

            if i % stream_interval == 0 or i == req.max_tokens - 1 or stopped:
                if echo:
                    tmp_output_ids = output_ids
                    rfind_start = input_length
                else:
                    tmp_output_ids = output_ids[input_length:]
                    rfind_start = 0

                output = self.tokenizer.decode(
                    tmp_output_ids,
                    skip_special_tokens=True,
                    spaces_between_special_tokens=False,
                )

                partially_stopped = False
                if stop_token:
                    if isinstance(stop_token, str):
                        pos = output.rfind(stop_token, rfind_start)
                        if pos != -1:
                            output = output[:pos]
                            stopped = True
                        else:
                            partially_stopped = partial_stop(output, stop_token)
                    elif isinstance(stop_token, Iterable):
                        for each_stop in stop_token:
                            pos = output.rfind(each_stop, rfind_start)
                            if pos != -1:
                                output = output[:pos]
                                stopped = True
                                break
                            else:
                                partially_stopped = partial_stop(output, each_stop)
                                if partially_stopped:
                                    break
                    else:
                        raise ValueError("Invalid stop field type.")

                # prevent yielding partial stop sequence
                if not partially_stopped:
                    pass

            if stopped:
                break

        # finish stream event, which contains finish reason
        if i == req.max_tokens - 1:
            finish_reason = "length"
        elif stopped:
            finish_reason = "stop"
        else:
            finish_reason = None

        yield resp_cls.from_message(
            message=output,
            model=self.model_name,
            finish_reason=finish_reason,
            prompt_token=input_length,
            completion_token=i,
        )

        # clean
        del past_key_values, out
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
